% In the name of Allah, The Rahman, The Raheem%
% MSABMPv3 - Function for executing the multiple measurement vectors (MMV)
% version of the Support Agnostic Bayesian Matching Pursuit (SABMP) algorithm.
%
% This function implements the MMV version of the SABMP algorithm described 
% in the following papers
% ---------------------------------------------------------------------
% [1] M. Masood and T. Y. Al-Naffouri, "Sparse Reconstruction Using
% Distribution Agnostic Bayesian Matching Pursuit," IEEE Trans. Signal
% Process., vol. 61, no. 21, pp. 5298-5309, Nov. 2013.
% [2] T. Y. Al-Naffouri and Mudassir Masood, "Distribution agnostic 
% structured sparsity recovery algorithms." Systems, Signal Processing and 
% their Applications (WoSSPA), 2013 8th International Workshop on. IEEE, 2013.
% ---------------------------------------------------------------------
% If you use this code or any (modified) part of it in any 
% publication, please cite the above-mentioned papers.
%
% MSABMPv3 will find a high-probability set of sparse solutions "X"
% to the underdetermined system of linear equations
%                           Y = AX + W,
% where W is complex Gaussian, while X may be from any distribution.
%
% INPUTS:
% Y -       M x L dimension matrix of measurements, where L is the number
% of measurements vectors.
% A -       M x (LxN) dimension sensing matrix (you may have multiple (L)
% sensing matrices of size MxN each stacked together as one matrix A of
% size M x (LxN). A=[A_1 A_2 ... A_L]. Note that you have to supply A of
% this size even if you are using same sensing matix for all measurements.
% The columns of A are required to be normalized to unit norm.
% p1_est -      Prior probability of active taps
% sig2w_est -   Noise variance
%
% OUTPUTS:
% xmmsef -       N x L dimension matrix of MMSE estimates of L unknown
% signals from L measurements
% xmmse_star_collection -  Significant conditional MMSE estimates of X, given supports
% and Y
% psy_star -    Corresponding significant conditional mixture vector
% posterior probabilities, p(S|Y)
% nu_star -     corresponding dominant support selection metric values
%
% Coded by: Mudassir Masood and Tareq Y. Al-Naffouri, KAUST, Saudi Arabia.
% E-mail: mudassir.masood@kaust.edu.sa, tareq.alnaffouri@kaust.edu.sa
% Copyright (c) Mudassir Masood, Tareq Y. Al-Naffouri, 2013.

function [xmmsef, xmmse_star_collection, psy_star, nu_star] = MSABMPv3(Y, A, p1_est, sig2w_est)

%check for size compatibility
[rY,cY]=size(Y);
[rA, cA] = size(A);
if(rY ~= rA || rem(cA,cY)~=0)
    error('dimension mismatch between A and Y.');
end
M=rY;
L=cY;
N=cA/L;
%Allocate variables for storage
col_prod=zeros(N,N,L);
col_y_prod=zeros(N,L);
Ey1v=zeros(L,N);
eta_root=zeros(M,N,L);

nuxt_max=zeros(L,L);
nuxt_max_ind=zeros(L,L);
max_nuxt_root_ind=zeros(1,L);
max_j=zeros(1,L);
min_j=zeros(1,L);

c2=zeros(L,1);
root_const=zeros(L,1);
cltt=zeros(L,L);
nuxt_root=zeros(L,N);

%One time calculations. To be used for calculating E_y and E_psi
for i=1:L
    col_prod(:,:,i)=A(:,(i-1)*N+1:i*N)'*A(:,(i-1)*N+1:i*N);
    col_y_prod(:,i)=A(:,(i-1)*N+1:i*N)'*Y(:,i);
    Ey1v(i,:)=transpose(col_y_prod(:,i)./diag(col_prod(:,:,i)));
    eta_root(:,:,i)=A(:,(i-1)*N+1:i*N).*repmat(Ey1v(i,:),M,1);
end


p1=p1_est;
p1_old=p1_est;
sig2w=sig2w_est;
invsig2w=(1/(2*sig2w)); %we merge the factor of 2 within invsig2w to avoid extra multiplications in the iterations. therefore, it should be kept in mind that invsig2w is not just the inverse of noise variance, but inverse of the factor 2*sig2w appearing in our metric.


% Use refined values to perform greedy search
c3=log(p1/(1-p1))-log(2*pi)/2-log(sig2w_est)/2; %a difference of c3 occurs everytime there is a change in model vector
P = min(M, 1 + ceil(N*p1 + erfcinv(1e-1)*sqrt(2*N*p1*(1 - p1))));  % search length
nuxt=zeros(L,L,N);

xmmse=cell(P,L,L);
xmmse(:)={zeros(N,1)};

Eysv=zeros(1,N,L,L);
for i=1:L
    for j=1:L
        Eysv(:,:,i,j)=Ey1v(i,:);
    end
    c2(i)=-invsig2w*norm(Y(:,i))^2;
    root_const(i)=N*log(1-p1)+c2(i);
    for j=1:L
        cltt(i,j)=root_const(i);
    end
    nuxt_root(i,:)=root_const(i)+(sum(eta_root(:,:,i).*conj(eta_root(:,:,i)),1))*invsig2w;
    [~, max_nuxt_root_ind(i)] = max(nuxt_root(i,:));
end
nuxt_root = nuxt_root + c3-log(2*pi)/2-log(sig2w_est)/2;
max_ind=max_nuxt_root_ind;

for i=1:L
    for j=1:L
        nuxt(i,j,:)=nuxt_root(i,:);
    end
end

Eps_oldv=zeros(L,L,1,N);
nu=zeros(P,L,L);
T=cell(L,L);

% start greedy search
for p = 1:P
    
    Eysv_new=zeros(p+1,N,L,L);
    Eps_oldv_new=zeros(L,L,p,N);
    for i=1:L %for each of the observation i...
        for j=1:L %...perform L searches
            Eps_oldv_tmp=(squeeze(Eps_oldv(i,j,:,:)));
            if(p<=2)
                Eps_oldv_tmp=transpose(Eps_oldv_tmp);
            end

            [Eps_oldv_new(i,j,:,:), T{i,j}, nuxt(i,j,:), Eys_selectedv, Eysv_new(:,:,i,j), cltt(i,j), nu(p,i,j)]=...
                SABMP_core_function(Eps_oldv_tmp, T{i,j}, p, nuxt(i,j,:),max_ind(j), col_prod(:,:,i), Eysv(:,:,i,j), cltt(i,j),N, col_y_prod(:,i), c3, invsig2w);
            xmmse{p,i,j}(T{i,j}) = Eys_selectedv;
            psy_star=compute_posteriors_from_nuxt(nuxt(i,j,:));
            [nuxt_max(i,j), nuxt_max_ind(i,j)] = max(squeeze(nuxt(i,j,:)));
            nuxt_max(i,j)=nuxt(i,j,nuxt_max_ind(i,j));
        end
        [~,max_j(i)]=max(nuxt_max(i,:));
    end
    
    i=1:L;
    max_ind=nuxt_max_ind((max_j(i)-1)*L+i);
    
    Eysv=Eysv_new;
    Eps_oldv=Eps_oldv_new;
    T_tmp=T;
    Eps_oldv_tmp=Eps_oldv;
    nuxt_tmp=nuxt;
    Eysv_tmp=Eysv;
    cltt_tmp=cltt;
    
    for i=1:L
        for j=1:L
            Eps_oldv(i,j,:,:)=Eps_oldv_tmp(i,max_j(j),:,:);
            nuxt(i,j,:,:)=nuxt_tmp(i,max_j(j),:,:);
            Eysv(:,:,i,j)=Eysv_tmp(:,:,i,max_j(j));
            cltt(i,j)=cltt_tmp(i,max_j(j));
            T{i,j}=T_tmp{j,max_j(j)};
        end
    end
end

%keyboard
%% Calculate stuff for statistically significant hypotheses
xmmsef=zeros(N,L);
xmmse_star_collection = cell(1,L);
counter11=0;
for i=1:L
    
    nu_tmp=nu(:,i,:);
    [~,indx] = sort(nu_tmp(:), 'descend');
    nu_max = nu_tmp(indx(1));
    num=length(nu_tmp(:));
    nu_star = nu_tmp(indx(1:num));
    psy_star = exp(nu_star - nu_max)/sum(exp(nu_star - nu_max));    % posteriors
    
    xmmset=xmmse(:,i,:);
    xmmse_star = cell(1,num);           % mmse est cond on mixing params
    for k = 1:num
        counter11=counter11+1;
        xmmse_star{k} = xmmset{indx(k)};
        xmmse_star_collection{counter11} = xmmset{indx(k)};
    end
    xmmsef(:,i) = [xmmse_star{:}]*psy_star;  % approximate mmse estimate
end
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [Eps_oldv, T, nuxt, Eys_selectedv, Eysv, cltt, nu]=...
    SABMP_core_function(Eps_oldv, T, p, nuxt, nstar, col_prod, Eysv, cltt,N, col_y_prod, c3,invsig2w)


Epsv=zeros(p,N);
Eys_newv=zeros(p+1,N);
%[nustar] = max(nuxt);            % find best extension
%nu = nustar;                           % replace worst explored node...
nu=nuxt(nstar);
if (p > 1)
    %preparing variables to compute E_psi (called Epsv in the code)
    q_old=(col_prod(T,nstar))';
    f_old=1-q_old*Eps_oldv(:,nstar);
    c1v=(1/f_old)*(q_old*(Eps_oldv(:,1:N))-col_prod(nstar,1:N));
    T=[T nstar];
    
    %compute E_psi
    Epsv(:,1:N)=[(Eps_oldv(:,1:N))+Eps_oldv(:,nstar)*c1v; -c1v];
else
    T=nstar;
    Epsv=col_prod(nstar,1:N);
end

Eps_oldv=Epsv; %save Epsv to be used in the next iteration
Eys_selectedv=Eysv(:,nstar);

%preparing variables to compute E_y (called Eys_newv in the code)
qv=col_prod(T,1:N);
fv=(1-sum(conj(qv).*Epsv,1));
fv(fv<eps)=eps;
c2v=transpose((transpose(1./fv)).*((qv'*Eys_selectedv)-(col_y_prod(1:N))));

%compute E_y
Eys_newv(:,1:N)=[repmat(Eys_selectedv,1,N)+Epsv(:,1:N).*repmat(c2v,p,1); -c2v];

Eysv=Eys_newv; %store new E_y to be used in next iteration
cltt=cltt+c3; %a difference of c3 occurs everytime there is a change in model vector
nuxt=(invsig2w)*real(sum(conj([repmat(col_y_prod(T),1,N);transpose(col_y_prod)]).*Eys_newv))+cltt; %computing the metrics for next stage
nuxt(T) = -inf;
end


function psy_star=compute_posteriors_from_nuxt(nuxt)
[nu_max,~]=max(nuxt(:));
nu_star = nuxt(:);
%nu_star=nu_star/nu_max;
%nu_max=max(nu_star);
psy_star = exp(nu_star - nu_max)/sum(exp(nu_star - nu_max));    % posteriors
end
