% cs_alg:   Function for performing l_1 minimization based compressed sensing
%
% This function will find the least l_1 norm "x" to fit an underdetermined system of
% linear equations
%                           y = Ax + w,
% where w is complex Gaussian.
%
% SYNTAX: x_hat = cs_alg(y, A, epsilon, weight)
%
% Inputs:
% y -       Vector of observations of length M
% A -       M x N Measurement matrix
% epsilon -   Noise dependent parameter to control error in solution
% weight -    Weight vector of length N (if given weighted l_1 minimization is performed)
% Outputs:
% x_hat -    Recovered signal of length N
%
% Coded by: Ebrahim Al-Safadi, University of Southern California, CA., and Tareq Y. Al-Naffouri, KAUST, Saudi Arabia.
% E-mail: alsafadi@usc.edu, tareq.alnaffouri@kaust.edu.sa
% Copyright (c) Ebrahim Al-Safadi, Tareq Y. Al-Naffouri

function x_hat = cs_alg(y, A, epsilon, weight)

[A_rows, A_cols] = size(A);

if exist('weight','var')
    % Weighted CS Algorithm
    cvx_begin
        cvx_quiet(true);
        variable x_hat(A_cols) complex
        minimize (norm(x_hat.*weight, 1))
        subject to
            norm(y - A*x_hat, 2) <= epsilon;
    cvx_end
else
    % CS Algorithm
    cvx_begin
        cvx_quiet(true);
        variable x_hat(A_cols) complex
        minimize (norm(x_hat, 1))
        subject to
            norm(y - A*x_hat, 2) <= epsilon;
    cvx_end
end